import ansible.constants as AC
import battleschool.constants as C

# before callbacks import
AC.DEFAULT_CALLBACK_PLUGIN_PATH = C.DEFAULT_CALLBACK_PLUGIN_PATH

import ansible.playbook
import ansible.utils.template
from ansible import callbacks
from ansible import errors
from ansible import utils

from battleschool.__init__ import __version__
from battleschool.printing import *
from battleschool.source.git import Git
from battleschool.source.local import Local
from battleschool.source.url import Url

from copy import deepcopy
from tempfile import gettempdir
from sys import platform as _platform
from urlparse import urlparse

import json
import os
import platform
import sys
import tempfile

# TODO: verify environment: ansible

def getSourceHandlers():
    handlers = [Git, Local, Url]
    # TODO: auto load sources
    # for name, obj in inspect.getmembers(sourcepkg):
    #     if inspect.ismodule(obj):
    #         package = obj.__dict__['__package__']
    #         if package is not None and package.startswith('battleschool.source'):
    #             for srcName, srcObj in inspect.getmembers(obj):
    #                 if inspect.isclass(srcObj) and srcName != 'Source':
    #                     print "Loading %s/%s" % (srcName, srcObj)
    #                     handlers.append(srcObj)
    return handlers


def load_config_path(options, inventory, sshpass, sudopass):
    if options.config_file:
        config_file = options.config_file
        parse_result = urlparse(config_file)
        if parse_result.scheme:
            url_options = deepcopy(options)
            url_options.update_sources = True
            url_options.cache_dir = gettempdir()
            name = 'downloaded_config.yml'
            sources = {
                'url': [
                    {'name': name, 'url': config_file}
                ]
            }
            display(banner("Downloading config from url"))
            url = Url(url_options, sources)
            files = url.run(inventory, sshpass, sudopass)
            return files[0]
        else:
            return config_file

    return "%s/config.yml" % options.config_dir


def main(args, battleschool_dir=None):
    if not battleschool_dir:
        battleschool_dir = "%s/.battleschool" % os.environ['HOME']

    # TODO: make battle OO or more modular
    #-----------------------------------------------------------
    # make ansible defaults, battleschool defaults
    AC.DEFAULT_HOST_LIST = C.DEFAULT_HOST_LIST
    AC.DEFAULT_SUDO_FLAGS = C.DEFAULT_SUDO_FLAGS

    #-----------------------------------------------------------
    # create parser for CLI options
    usage = "%prog"
    parser = utils.base_parser(
        constants=AC,
        usage=usage,
        connect_opts=True,
        runas_opts=False,
        subset_opts=True,
        check_opts=True,
        diff_opts=True,
        output_opts=True
    )
    parser.version = "%s %s" % ("battleschool", __version__)
    parser.add_option('-e', '--extra-vars', dest="extra_vars", default=None,
                      help="set additional key=value variables from the CLI")
    # parser.add_option('--tags', dest='tags', default='all',
    #                   help="only run plays and tasks tagged with these values")
    parser.add_option('--syntax-check', dest='syntax', action='store_true',
                      help="do a playbook syntax check on the playbook, do not execute the playbook")
    parser.add_option('--list-tasks', dest='listtasks', action='store_true',
                      help="do list all tasks that would be executed")
    parser.add_option('--step', dest='step', action='store_true',
                      help="one-step-at-a-time: confirm each task before running")
    parser.add_option("-s", "--sudo", default=AC.DEFAULT_SUDO, action="store_true",
                      dest='sudo', help="run operations with sudo (nopasswd)")
    parser.add_option('--config-dir', dest='config_dir', default=None,
                      help="config directory for battleschool (default=%s)" % battleschool_dir)
    parser.add_option('--config-file', dest='config_file', default=None,
                      help="config file for battleschool (default=%s/%s)" % (battleschool_dir, "config.yml"))
    parser.add_option('-u', '--update-sources', dest='update_sources', default=False, action='store_true',
                      help="update playbooks from sources(git, url, etc...)")
    parser.add_option('--acquire-only', dest='acquire_only', default=False, action='store_true',
                      help="configure mac_pkg module to only aquire package (ie download only")

    options, args = parser.parse_args(args)
    # options.connection = 'local'

    playbooks_to_run = []  #[C.DEFAULT_PLAYBOOK]

    #-----------------------------------------------------------
    # setup inventory
    inventory = ansible.inventory.Inventory(options.inventory)
    inventory.subset(options.subset)
    if len(inventory.list_hosts()) == 0:
        raise errors.AnsibleError("provided hosts list is empty")

    #-----------------------------------------------------------
    # setup default options
    sshpass = None
    sudopass = None
    options.remote_user = AC.DEFAULT_REMOTE_USER
    if not options.listhosts and not options.syntax and not options.listtasks:
        options.ask_pass = AC.DEFAULT_ASK_PASS
        options.ask_sudo_pass = options.ask_sudo_pass or AC.DEFAULT_ASK_SUDO_PASS
        passwds = utils.ask_passwords(ask_pass=options.ask_pass, ask_sudo_pass=options.ask_sudo_pass)
        sshpass = passwds[0]
        sudopass = passwds[1]
        # if options.sudo_user or options.ask_sudo_pass:
        #     options.sudo = True
        options.sudo_user = AC.DEFAULT_SUDO_USER
    if options.extra_vars and options.extra_vars[0] in '[{':
        extra_vars = utils.json_loads(options.extra_vars)
    else:
        extra_vars = utils.parse_kv(options.extra_vars)
    only_tags = None  # options.tags.split(",")

    #-----------------------------------------------------------
    # setup config_dir and battleschool_dir
    if options.config_dir:
        battleschool_dir = options.config_dir
    else:
        options.config_dir = battleschool_dir

    #-----------------------------------------------------------
    # setup module_path
    if options.module_path is None:
        options.module_path = AC.DEFAULT_MODULE_PATH

    if options.module_path is None:
        options.module_path = C.DEFAULT_MODULE_PATH

    if C.DEFAULT_MODULE_PATH not in options.module_path:
        options.module_path = "%s:%s" % (C.DEFAULT_MODULE_PATH, options.module_path)

    #-----------------------------------------------------------
    # parse config data
    config_path = load_config_path(options, inventory, sshpass, sudopass)
    if os.path.exists(config_path) and os.path.isfile(config_path):
        config_data = utils.parse_yaml_from_file(config_path)
    else:
        config_data = {}

    #-----------------------------------------------------------
    # set config_dir
    if "cache_dir" in config_data:
        options.cache_dir = os.path.expanduser(config_data["cache_dir"])
    elif _platform == "darwin":  # OS X
        options.cache_dir = os.path.expanduser("~/Library/Caches/battleschool")
    else:
        options.cache_dir = "%s/cache" % battleschool_dir

    os.environ["BATTLESCHOOL_CACHE_DIR"] = options.cache_dir

    #-----------------------------------------------------------
    # setup extra_vars for later use
    if extra_vars is None:
        extra_vars = dict()

    extra_vars['battleschool_config_dir'] = battleschool_dir
    extra_vars['battleschool_cache_dir'] = options.cache_dir
    extra_vars['mac_pkg_acquire_only'] = options.acquire_only

    #-----------------------------------------------------------
    # set mac_version for extra_vars
    if _platform == "darwin":
        mac_version = platform.mac_ver()[0].split(".")
        extra_vars['mac_version'] = mac_version
        extra_vars['mac_major_minor_version'] = "%s.%s" % (mac_version[0], mac_version[1])

    #-----------------------------------------------------------
    # serialize extra_vars since there is now way to pass data
    # to a module without modifying every playbook
    tempdir = tempfile.gettempdir()
    extra_vars_path = os.path.join(tempdir, "battleschool_extra_vars.json")
    with open(extra_vars_path, 'w') as f:
        f.write(json.dumps(extra_vars))

    #-----------------------------------------------------------
    # setup and run source handlers
    handlers = getSourceHandlers()

    if 'sources' in config_data and config_data['sources']:
        sources = config_data['sources']
        display(banner("Updating sources"))
        for handler in handlers:
            source = handler(options, sources)
            playbooks = source.run(inventory, sshpass, sudopass)
            for playbook in playbooks:
                playbooks_to_run.append(playbook)
    else:
        display(banner("No sources to update"))

    #-----------------------------------------------------------
    # validate playbooks
    for playbook in playbooks_to_run:
        if not os.path.exists(playbook):
            raise errors.AnsibleError("the playbook: %s could not be found" % playbook)
        if not os.path.isfile(playbook):
            raise errors.AnsibleError("the playbook: %s does not appear to be a file" % playbook)

    #-----------------------------------------------------------
    # run all playbooks specified from config
    for playbook in playbooks_to_run:
        stats = callbacks.AggregateStats()

        # let inventory know which playbooks are using so it can know the basedirs
        inventory.set_playbook_basedir(os.path.dirname(playbook))

        runner_cb = BattleschoolRunnerCallbacks()
        playbook_cb = BattleschoolCallbacks()
        #TODO: option to use default callbacks
        # runner_cb = callbacks.PlaybookRunnerCallbacks(stats, verbose=utils.VERBOSITY)
        # playbook_cb = callbacks.PlaybookCallbacks(verbose=utils.VERBOSITY)

        if options.step:
            playbook_cb.step = options.step

        pb = ansible.playbook.PlayBook(
            playbook=playbook,
            module_path=options.module_path,
            inventory=inventory,
            forks=options.forks,
            remote_user=options.remote_user,
            remote_pass=sshpass,
            callbacks=playbook_cb,
            runner_callbacks=runner_cb,
            stats=stats,
            timeout=options.timeout,
            transport=options.connection,
            sudo=options.sudo,
            sudo_user=options.sudo_user,
            sudo_pass=sudopass,
            extra_vars=extra_vars,
            private_key_file=options.private_key_file,
            only_tags=only_tags,
            check=options.check,
            diff=options.diff
        )

        if options.listhosts or options.listtasks:
            print ''
            print 'playbook: %s' % playbook
            print ''
            playnum = 0
            for (play_ds, play_basedir) in zip(pb.playbook, pb.play_basedirs):
                playnum += 1
                play = ansible.playbook.Play(pb, play_ds, play_basedir)
                label = play.name
                if options.listhosts:
                    hosts = pb.inventory.list_hosts(play.hosts)
                    print '  play #%d (%s): host count=%d' % (playnum, label, len(hosts))
                    for host in hosts:
                        print '    %s' % host
                if options.listtasks:
                    matched_tags, unmatched_tags = play.compare_tags(pb.only_tags)
                    unmatched_tags.discard('all')
                    unknown_tags = set(pb.only_tags) - (matched_tags | unmatched_tags)
                    if unknown_tags:
                        continue
                    print '  play #%d (%s): task count=%d' % (playnum, label, len(play.tasks()))
                    for task in play.tasks():
                        if set(task.tags).intersection(pb.only_tags):
                            if getattr(task, 'name', None) is not None:
                                # meta tasks have no names
                                print '    %s' % task.name
                print ''
            continue

        if options.syntax:
            # if we've not exited by now then we are fine.
            print 'Playbook Syntax is fine'
            return 0

        failed_hosts = []

        try:

            pb.run()

            hosts = sorted(pb.stats.processed.keys())
            # display(callbacks.banner("PLAY RECAP"))
            playbook_cb.on_stats(pb.stats)

            for host in hosts:
                smry = pb.stats.summarize(host)
                if smry['unreachable'] > 0 or smry['failures'] > 0:
                    failed_hosts.append(host)

            if len(failed_hosts) > 0:
                filename = pb.generate_retry_inventory(failed_hosts)
                if filename:
                    display("           to retry, use: --limit @%s\n" % filename)

            for host in hosts:
                smry = pb.stats.summarize(host)
                print_stats(host, smry)

            # print ""
            if len(failed_hosts) > 0:
                return 2

        except errors.AnsibleError, e:
            display("ERROR: %s" % e, color='red')
            return 1

    if not playbooks_to_run:
        display("\tWARNING: no playbooks run!", color='yellow')

    os.remove(extra_vars_path)
    display(banner("Battleschool completed"))
    # TODO: aggregate stats across playbook runs



__author__ = 'spencergibb'


from . import Source


class Local(Source):
    """git source handler.
    """

    def type(self):
        return 'local'

    def run(self, inventory, sshpass, sudopass):
        playbooks = []

        if self.has_source_type():
            for source in self.sources[self.type()]:
                playbook = "%s/playbooks/%s" % (self.options.config_dir, source)
                self.add_playbook(playbooks, playbook)

        return playbooks

    def module_args(self, source):
        pass



__author__ = 'spencergibb'


from . import Source


class Git(Source):
    """git source handler.
    """

    def type(self):
        return 'git'

    def dest_dir(self, source):
        dest_dir = "%s/%s" % (self.options.cache_dir, source['name'])
        return dest_dir

    def module_args(self, source):
        force = "no"
        update = "no"

        if self.options.update_sources:
            update = "yes"
            force = "yes"

        module_args = "repo=%s dest=%s force=%s update=%s " % \
                      (source['repo'], self.dest_dir(source), force, update)
        return module_args





class CallbackModule(object):

    def display(self, event, *args):

        play = getattr(self, 'play', None)
        task = getattr(self, 'task', None)

        if play is not None:
            playbook = play.playbook.filename
        else:
            playbook = ""

        if task is not None:
            task_name = task.name
        else:
            task_name = ""

        print("%25s: play = %s, task = %s, args = %s" % (event, playbook, task_name, args))

    def on_any(self, *args, **kwargs):
        # play = getattr(self, 'play', None)
        # task = getattr(self, 'task', None)
        # print "on_any: play = %s, task = %s, args = %s, kwargs = %s" % (play, task, args, kwargs)
        pass

    def runner_on_failed(self, host, res, ignore_errors=False):
        # self.display('runner_on_failed', host, res, ignore_errors)
        pass

    def runner_on_ok(self, host, res):
        # self.display('runner_on_ok', host, res)
        pass

    def runner_on_error(self, host, msg):
        # self.display('runner_on_error', host, msg)
        pass

    def runner_on_skipped(self, host, item=None):
        # self.display('runner_on_skipped', host, item)
        pass

    def runner_on_unreachable(self, host, res):
        # self.display('runner_on_unreachable', host, res)
        pass

    def runner_on_no_hosts(self):
        # self.display('runner_on_no_hosts')
        pass

    def runner_on_async_poll(self, host, res, jid, clock):
        self.display('runner_on_async_poll', host, res, jid, clock)

    def runner_on_async_ok(self, host, res, jid):
        self.display('runner_on_async_ok', host, res, jid)

    def runner_on_async_failed(self, host, res, jid):
        self.display('runner_on_async_failed', host, res, jid)

    def playbook_on_start(self):
        # self.display('playbook_on_start')
        pass

    def playbook_on_notify(self, host, handler):
        self.display('playbook_on_notify', host, handler)

    def playbook_on_no_hosts_matched(self):
        self.display('playbook_on_no_hosts_matched')

    def playbook_on_no_hosts_remaining(self):
        # self.display('playbook_on_no_hosts_remaining')
        pass

    def playbook_on_task_start(self, name, is_conditional):
        # self.display('playbook_on_task_start', name, is_conditional)
        pass

    def playbook_on_vars_prompt(self, varname, private=True, prompt=None, encrypt=None, confirm=False, salt_size=None, salt=None, default=None):
        self.display('playbook_on_vars_prompt', varname, private, prompt, encrypt, confirm, salt_size, salt, default)

    def playbook_on_setup(self):
        # self.display('playbook_on_setup')
        pass

    def playbook_on_import_for_host(self, host, imported_file):
        self.display('playbook_on_import_for_host', host, imported_file)

    def playbook_on_not_import_for_host(self, host, missing_file):
        self.display('playbook_on_not_import_for_host', host, missing_file)

    def playbook_on_play_start(self, pattern):
        # self.display('playbook_on_play_start', pattern)
        pass

    def playbook_on_stats(self, stats):
        # self.display('playbook_on_stats', stats)
        pass




__author__ = 'spencergibb'


from . import Source


class Url(Source):
    """url source handler.
    """

    def type(self):
        return 'url'

    def module_name(self):
        return 'get_url'

    def dest_dir(self, source):
        return self.options.cache_dir

    def module_args(self, source):
        force = "no"

        if self.options.update_sources:
            force = "yes"

        module_args = "url=%s dest=%s/%s force=%s validate_certs=no " % \
                      (source["url"], self.dest_dir(source), source["name"], force)

        if "playbooks" not in source:
            source["playbooks"] = []

        source["playbooks"].append(source["name"])
        return module_args



import os
import sys


def get_config(env_var, default):
    """ return a configuration variable with casting """
    if env_var is not None:
        value = os.environ.get(env_var, None)
        if value is not None:
            return value
    return default


# Needed so the RPM can call setup.py and have modules land in the
# correct location. See #1277 for discussion
if getattr(sys, "real_prefix", None):
    DIST_MODULE_PATH = os.path.join(sys.prefix, 'share/battleschool/')
else:
    DIST_MODULE_PATH = '/usr/share/battleschool/'

DEFAULT_MODULE_PATH = get_config('BATTLESCHOOL_LIBRARY', os.path.join(DIST_MODULE_PATH, 'library'))
DEFAULT_HOST_LIST = os.path.expanduser(get_config('BATTLESCHOOL_HOSTS', os.path.join(DIST_MODULE_PATH, 'defaults', 'hosts')))
DEFAULT_PLAYBOOK = os.path.expanduser(get_config('BATTLESCHOOL_PLAYBOOK', os.path.join(DIST_MODULE_PATH, 'defaults', 'battleschool.yml')))
DEFAULT_CALLBACK_PLUGIN_PATH = os.path.expanduser(get_config('BATTLESCHOOL_CALLBACK_PLUGINS', os.path.join(DIST_MODULE_PATH, 'callback_plugins')))
DEFAULT_SUDO_FLAGS = get_config('BATTLESCHOOL_SUDO_FLAGS', '-E')




from battleschool.printing import BattleschoolRunnerCallbacks

__author__ = 'spencergibb'


import abc
import os
import sys

from ansible import errors
from ansible.callbacks import display
from ansible.runner import Runner


class Source(object):
    """Base class for source handlers.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, options, sources):
        self.options = options
        if (sources):
            self.sources = sources
        else:
            self.sources = {}
        return

    @abc.abstractmethod
    def type(self):
        pass

    @abc.abstractmethod
    def module_args(self, source):
        """Override to do something useful.
        """

    def has_source_type(self):
        return self.type() in self.sources and self.sources[self.type()] is not None

    def dest_dir(self, source):
        dest_dir = "%s/%s" % (self.options.cache_dir, source['name'])
        return dest_dir

    def module_name(self):
        return self.type()

    def add_playbook(self, playbooks, playbook):
        if os.path.exists(playbook) and os.path.isfile(playbook):
            playbooks.append(playbook)

    def run_module(self, inventory, source, sshpass, sudopass):
        runner_cb = BattleschoolRunnerCallbacks()
        runner_cb.options = self.options
        runner_cb.options.module_name = self.module_name()
        module_args = self.module_args(source)
        #TODO: get workstation from options
        runner = Runner(
            pattern='workstation',
            module_name=self.module_name(),
            module_path=self.options.module_path,
            module_args=module_args,
            inventory=inventory,
            callbacks=runner_cb,
            timeout=self.options.timeout,
            transport=self.options.connection,
            #sudo=self.options.sudo,
            sudo=False,
            sudo_user=self.options.sudo_user,
            sudo_pass=sudopass,
            check=self.options.check,
            diff=self.options.diff,
            private_key_file=self.options.private_key_file,
            remote_user=self.options.remote_user,
            remote_pass=sshpass,
            forks=self.options.forks
        )
        try:
            results = runner.run()
            for result in results['contacted'].values():
                if 'failed' in result or result.get('rc', 0) != 0:
                    display("ERROR: failed source type (%s) '%s': %s" % (self.type(), module_args, result['msg']),
                            stderr=True, color='red')
                    sys.exit(2)
            if results['dark']:
                display("ERROR: failed source type (%s) '%s': DARK" % (self.type(), module_args),
                        stderr=True, color='red')
                sys.exit(2)
        except errors.AnsibleError, e:
            # Generic handler for ansible specific errors
            display("ERROR: %s" % str(e), stderr=True, color='red')
            sys.exit(1)

    def run(self, inventory, sshpass, sudopass):
        playbooks = []
        if self.has_source_type():
            for source in self.sources[self.type()]:
                # print source
                self.run_module(inventory, source, sshpass, sudopass)

                source_playbooks = ["local.yml"]

                #add other playbooks relative to dest_dir from config.yml
                if "playbooks" in source and source['playbooks'] is not None:
                    for playbook in source['playbooks']:
                        # support for single level of directories in a playbook repo
                        if type(playbook) is dict:
                            for dir in playbook:
                                names = playbook[dir]
                                for name in names:
                                    playbook_name = "%s/%s" % (dir, name)
                                    if playbook_name not in source_playbooks:
                                        source_playbooks.append(playbook_name)
                        elif playbook not in source_playbooks:
                            source_playbooks.append(playbook)

                for playbook_name in source_playbooks:
                    suffix = ""

                    if not playbook_name.endswith(".yml") and not playbook_name.endswith(".yaml"):
                        suffix = ".yml"

                    playbook = "%s/%s%s" % (self.dest_dir(source), playbook_name, suffix)
                    self.add_playbook(playbooks, playbook)

        return playbooks



#!/usr/bin/env python

import os
import sys
from glob import glob

sys.path.insert(0, os.path.abspath('lib'))
from battleschool import __version__, __author__
from distutils.core import setup

# find library modules
from battleschool.constants import DIST_MODULE_PATH

long_description = """
Development environment provisioning using ansible (http://docs.ansible.com),
ala boxen (http://boxen.github.com/) which uses puppet (http://puppetlabs.com/puppet/what-is-puppet) and
kitchenplan (https://github.com/kitchenplan/kitchenplan) which uses chef (http://docs.opscode.com/)
Built on and for macs, but should be usable on Linux
"""

share_path = "./share/"
files = os.listdir(share_path)
data_files = []
for i in files:
    if os.path.isdir(os.path.join(share_path, i)):
        data_files.append((DIST_MODULE_PATH + i, glob(share_path + i + '/*')))

    #if os.path.isfile(os.path.join(share_path, i)):
    #    data_files.append((DIST_MODULE_PATH, share_path + i))

setup(name='battleschool',
      version=__version__,
      description='simple dev box provisioning',
      long_description=long_description,
      author=__author__,
      author_email='spencer@gibb.us',
      url='http://spencer.gibb.us',
      download_url='https://github.com/spencergibb/battleschool/releases',
      license='Apache License, Version 2.0',
      install_requires=['ansible', 'jinja2', 'pyyaml'],  # added jinja2 and pyyaml to fix installs under homebrew pip
      classifiers=[
          "Development Status :: 3 - Alpha",
          "Environment :: Console",
          "Environment :: MacOS X",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: Apache Software License",
          "Programming Language :: Python :: 2.7",
          "Topic :: System :: Installation/Setup"
      ],
      keywords="provisioning setup install",
      package_dir={'battleschool': 'lib/battleschool'},
      packages=[
          'battleschool',
          'battleschool.source',
      ],
      scripts=[
          'bin/battle'
      ],
      data_files=data_files
)



from ansible.callbacks import call_callback_module
from ansible.callbacks import display
from ansible.callbacks import DefaultRunnerCallbacks
from ansible.color import ANSIBLE_COLOR, stringc

import pprint

pp = pprint.PrettyPrinter()


def colorize(lead, num, color):
    """ Print 'lead' = 'num' in 'color' """
    if num != 0 and ANSIBLE_COLOR and color is not None:
        return "%s%s%s" % (stringc(lead, color), stringc("=", color), stringc(str(num), color))
    else:
        return "%s=%s" % (lead, str(num))


def hostcolor(host, stats, color=True):
    if ANSIBLE_COLOR and color:
        if stats['failures'] != 0 or stats['unreachable'] != 0:
            return "%s" % stringc(host, 'red')
        elif stats['changed'] != 0:
            return "%s" % stringc(host, 'yellow')
        else:
            return "%s" % stringc(host, 'green')
    return "%s" % host


def print_stats(host, smry):
    ok = smry['ok']
    changed = smry['changed']
    unreachable = smry['unreachable']
    failures = smry['failures']

    if unreachable > 0 or failures > 0:
        status = stringc("FAILED", "red")
    else:
        status = stringc("OK", "bright green")

    pattern = "\tPlaybook %s, %s, %s, %s, %s"

    display(pattern % (status,
        # hostcolor(host, smry, False),
        colorize('ok', ok, 'bright green'),
        colorize('changed', changed, 'yellow'),
        colorize('unreachable', unreachable, 'red'),
        colorize('failed', failures, 'red')),
            screen_only=True
    )

    display(pattern % (status,
        # hostcolor(host, smry, False),
        colorize('ok', ok, None),
        colorize('changed', changed, None),
        colorize('unreachable', unreachable, None),
        colorize('failed', failures, None)),
            log_only=True
    )


def banner(msg):
    #TODO: configurable size orig 78
    width = 110 - len(msg)
    if width < 3:
        width = 3
    filler = "#" * width
    return "## %s %s " % (msg, filler)


class BattleschoolRunnerCallbacks(DefaultRunnerCallbacks):
    """ callbacks for use by battles """

    def get_play(self):
        if self.runner:
            return self.runner
        if self.task:
            return self.task
        return self.play

    def get_name(self):
        if self.runner:
            return self.get_play().module_name
        return self.get_play().name

    def on_failed(self, host, res, ignore_errors=False):
        if not ignore_errors:
            display("\tTask FAILED: %s %s" % (self.get_name(), res['msg']), color="red")
        super(BattleschoolRunnerCallbacks, self).on_failed(host, res, ignore_errors=ignore_errors)

    def on_ok(self, host, res):
        if 'msg' in res:
            msg = ": %s" % res['msg']

            if 'item' in res:
                msg = "%s => item=%s" % (msg, res['item'])
        else:
            msg = ''

        try:
            if self.get_play():
                display("\tTask OK: %s%s" % (self.get_name(), msg))
        except AttributeError:
            invocation = res['invocation']
            msg = invocation['module_args']
            module_name = invocation['module_name']

            #TODO: move this?
            if module_name == 'git':
                args = dict()
                tokens = invocation['module_args'].split()
                for token in tokens:
                    pair = token.split("=")[:2]
                    args[pair[0]] = pair[1]
                msg = args['repo']

            display("\tTask OK: %s, changed=%s %s" % (module_name, res['changed'], msg))
        super(BattleschoolRunnerCallbacks, self).on_ok(host, res)

    def on_unreachable(self, host, results):
        item = None
        if type(results) == dict:
            item = results.get('item', None)
        if item:
            msg = "\tFailed Task: %s => (item=%s) => %s" % (host, item, results)
        else:
            msg = "\tFatal Task: %s => %s" % (host, results)
        display(msg, color='red', runner=self.runner)
        super(BattleschoolRunnerCallbacks, self).on_unreachable(host, results)

    def on_skipped(self, host, item=None):
        display("\tTask skipped: %s" % self.get_name(), color="yellow")
        super(BattleschoolRunnerCallbacks, self).on_skipped(host, item)

    def on_error(self, host, err):
        display("\tTask ERROR: %s%s" % (self.get_name(), err), color="red")
        super(BattleschoolRunnerCallbacks, self).on_error(host, err)

    def on_no_hosts(self):
        display("\tTask NO HOSTS: %s" % self.get_name(), color="red")
        super(BattleschoolRunnerCallbacks, self).on_no_hosts()

    def on_async_poll(self, host, res, jid, clock):
        super(BattleschoolRunnerCallbacks, self).on_async_poll(host, res, jid, clock)

    def on_async_ok(self, host, res, jid):
        super(BattleschoolRunnerCallbacks, self).on_async_ok(host, res, jid)

    def on_async_failed(self, host, res, jid):
        super(BattleschoolRunnerCallbacks, self).on_async_failed(host,res,jid)

    def on_file_diff(self, host, diff):
        super(BattleschoolRunnerCallbacks, self).on_file_diff(host, diff)


class BattleschoolCallbacks(object):
    """ used by battle for playbooks """

    def __init__(self, verbose=False):

        self.verbose = verbose

    def on_start(self):
        call_callback_module('playbook_on_start')

    def on_notify(self, host, handler):
        call_callback_module('playbook_on_notify', host, handler)

    def on_no_hosts_matched(self):
        call_callback_module('playbook_on_no_hosts_matched')

    def on_no_hosts_remaining(self):
        display("\tFailed playbook: %s" % self.playbook.filename, color='bright red')
        call_callback_module('playbook_on_no_hosts_remaining')

    def on_task_start(self, name, is_conditional):
        call_callback_module('playbook_on_task_start', name, is_conditional)

    def on_vars_prompt(self, varname, private=True, prompt=None, encrypt=None, confirm=False, salt_size=None, salt=None, default=None):
        call_callback_module( 'playbook_on_vars_prompt', varname, private=private, prompt=prompt,
                              encrypt=encrypt, confirm=confirm, salt_size=salt_size, salt=None, default=default )

    def on_setup(self):
        call_callback_module('playbook_on_setup')

    def on_import_for_host(self, host, imported_file):
        call_callback_module('playbook_on_import_for_host', host, imported_file)

    def on_not_import_for_host(self, host, missing_file):
        call_callback_module('playbook_on_not_import_for_host', host, missing_file)

    def on_play_start(self, pattern):
        display(banner("Executing playbook %s" % self.playbook.filename), color="bright blue")
        call_callback_module('playbook_on_play_start', pattern)

    def on_stats(self, stats):
        call_callback_module('playbook_on_stats', stats)
